@@ -81,10 +81,13 @@ struct osmo_stream_srv_link {
 	uint16_t		port;
 	int			sk_domain;
 	int			sk_type;
+	int			sk_prio; /* socket priority, SO_PRIORITY, default=0=unset */
 	uint16_t		proto;
+	uint8_t			ip_dscp; /* IP Differentiated services, 0..63, default=0=unset */
 	osmo_stream_srv_link_accept_cb_t accept_cb;
 	void			*data;
 	int			flags;
+	unsigned int		tx_queue_max_length; /* Max amount of msgbs which can be enqueued */
 	struct osmo_sock_init2_multiaddr_pars ma_pars;
 };
 
@@ -148,6 +151,24 @@ static int osmo_stream_srv_link_ofd_cb(struct osmo_fd *ofd, unsigned int what)
 			goto error_close_socket;
 	}
 
+	if (link->ip_dscp > 0) {
+		ret = osmo_sock_set_dscp(sock_fd, link->ip_dscp);
+		if (ret < 0) {
+			LOGSLNK(link, LOGL_ERROR, "set_ip_dscp(%u): failed setsockopt err=%d\n",
+				link->ip_dscp, errno);
+			goto error_close_socket;
+		}
+	}
+
+	if (link->sk_prio > 0) {
+		ret = osmo_sock_set_priority(sock_fd, link->sk_prio);
+		if (ret < 0) {
+			LOGSLNK(link, LOGL_ERROR, "set_priority(%d): failed setsockopt err=%d\n",
+				link->sk_prio, errno);
+			goto error_close_socket;
+		}
+	}
+
 	if (!link->accept_cb) {
 		ret = -ENOTSUP;
 		goto error_close_socket;
@@ -183,6 +204,7 @@ struct osmo_stream_srv_link *osmo_stream_srv_link_create(void *ctx)
 	link->sk_domain = AF_UNSPEC;
 	link->sk_type = SOCK_STREAM;
 	link->proto = IPPROTO_TCP;
+	link->tx_queue_max_length = 1024; /* Default tx queue size, msgbs. */
 	osmo_fd_setup(&link->ofd, -1, OSMO_FD_READ | OSMO_FD_WRITE, osmo_stream_srv_link_ofd_cb, link, 0);
 
 	link->ma_pars.sctp.version = 0;
@@ -224,6 +246,33 @@ void osmo_stream_srv_link_set_nodelay(struct osmo_stream_srv_link *link, bool no
 		link->flags &= ~OSMO_STREAM_SRV_F_NODELAY;
 }
 
+/*! Set the priority value of the stream socket.
+ *  Setting this will automatically set the socket priority
+ *  option on any socket established via this server link, before
+ *  calling the accept_cb().
+ *  \param[in] link server link whose sockets are to be configured
+ *  \param[in] sk_prio priority value. Values outside 0..6 require CAP_NET_ADMIN.
+ *  \return negative on error, 0 on success
+ */
+int osmo_stream_srv_link_set_priority(struct osmo_stream_srv_link *link, int sk_prio)
+{
+	link->sk_prio = sk_prio;
+	return 0;
+}
+
+/*! Set the DSCP (differentiated services code point) of the stream socket.
+ *  Setting this  will automatically set the IP DSCP option on any socket on any
+ *  socket established via this server link, before calling the accept_cb().
+ *  \param[in] link server link whose sockets are to be configured
+ *  \param[in] ip_dscp DSCP value. Value range 0..63.
+ *  \return negative on error, 0 on success
+ */
+int osmo_stream_srv_link_set_ip_dscp(struct osmo_stream_srv_link *link, uint8_t ip_dscp)
+{
+	link->ip_dscp = ip_dscp;
+	return 0;
+}
+
 /*! Set the local address to which we bind.
  *  Any changes to this setting will only become active upon next (re)connect.
  *  \param[in] link Stream Server Link to modify
@@ -347,6 +396,18 @@ void *osmo_stream_srv_link_get_data(struct osmo_stream_srv_link *link)
 	return link->data;
 }
 
+/*! Set the maximum length queue of the stream servers accepted and allocated from this server link.
+ *  \param[in] link Stream Server Link to modify
+ *  \param[in] size maximum amount of msgbs which can be queued in the internal tx queue.
+ *  \returns 0 on success, negative on error.
+ *
+ *  The maximum length queue default value is 1024 msgbs. */
+int osmo_stream_srv_link_set_tx_queue_max_length(struct osmo_stream_srv_link *link, unsigned int size)
+{
+	link->tx_queue_max_length = size;
+	return 0;
+}
+
 /* Similar to osmo_sock_multiaddr_get_name_buf(), but aimed at listening sockets (only local part): */
 static char *get_local_sockname_buf(char *buf, size_t buf_len, const struct osmo_stream_srv_link *link)
 {
@@ -449,6 +510,8 @@ void osmo_stream_srv_link_set_accept_cb(struct osmo_stream_srv_link *link,
  *  \param[in] link Stream Server Link */
 void osmo_stream_srv_link_destroy(struct osmo_stream_srv_link *link)
 {
+	if (!link)
+		return;
 	osmo_stream_srv_link_close(link);
 	talloc_free(link);
 }
@@ -596,10 +659,13 @@ struct osmo_stream_srv {
 		struct osmo_fd			ofd;
 		struct osmo_io_fd		*iofd;
 	};
-	struct llist_head		tx_queue;
+	struct llist_head		tx_queue; /* osmo_ofd mode (only): Queue of msgbs */
+	unsigned int			tx_queue_count; /* osmo_ofd mode (only): Current amount of msgbs queued */
 	osmo_stream_srv_closed_cb_t	closed_cb;
 	osmo_stream_srv_read_cb_t	read_cb;
 	osmo_stream_srv_read_cb2_t	iofd_read_cb;
+	osmo_stream_srv_segmentation_cb_t segmentation_cb;
+	osmo_stream_srv_segmentation_cb2_t segmentation_cb2;
 	void				*data;
 	int				flags;
 };
@@ -722,12 +788,11 @@ static void osmo_stream_srv_write(struct osmo_stream_srv *conn)
 	struct msgb *msg;
 	int ret;
 
-	if (llist_empty(&conn->tx_queue)) {
+	msg = msgb_dequeue_count(&conn->tx_queue, &conn->tx_queue_count);
+	if (!msg) { /* done, tx_queue empty */
 		osmo_fd_write_disable(&conn->ofd);
 		return;
 	}
-	msg = llist_first_entry(&conn->tx_queue, struct msgb, list);
-	llist_del(&msg->list);
 
 	LOGSSRV(conn, LOGL_DEBUG, "sending %u bytes of data\n", msg->len);
 
@@ -764,6 +829,7 @@ static void osmo_stream_srv_write(struct osmo_stream_srv *conn)
 		/* Update msgb and re-add it at the start of the queue: */
 		msgb_pull(msg, ret);
 		llist_add(&msg->list, &conn->tx_queue);
+		conn->tx_queue_count++;
 		return;
 	}
 
@@ -773,6 +839,7 @@ static void osmo_stream_srv_write(struct osmo_stream_srv *conn)
 		if (err == EAGAIN) {
 			/* Re-add at the start of the queue to re-attempt: */
 			llist_add(&msg->list, &conn->tx_queue);
+			conn->tx_queue_count++;
 			return;
 		}
 		msgb_free(msg);
@@ -876,6 +943,7 @@ osmo_stream_srv_create2(void *ctx, struct osmo_stream_srv_link *link, int fd, vo
 
 	conn->mode = OSMO_STREAM_MODE_OSMO_IO;
 	conn->srv = link;
+	conn->data = data;
 
 	osmo_sock_get_name_buf(conn->sockname, sizeof(conn->sockname), fd);
 
@@ -892,7 +960,8 @@ osmo_stream_srv_create2(void *ctx, struct osmo_stream_srv_link *link, int fd, vo
 		talloc_free(conn);
 		return NULL;
 	}
-	conn->data = data;
+
+	osmo_iofd_set_txqueue_max_length(conn->iofd, conn->srv->tx_queue_max_length);
 
 	if (osmo_iofd_register(conn->iofd, fd) < 0) {
 		LOGSSRV(conn, LOGL_ERROR, "could not register FD %d\n", fd);
@@ -973,7 +1042,37 @@ osmo_stream_srv_set_data(struct osmo_stream_srv *conn,
 	conn->data = data;
 }
 
+/* Callback from iofd, forward to stream_srv user: */
+static int stream_srv_iofd_segmentation_cb2(struct osmo_io_fd *iofd, struct msgb *msg)
+{
+	struct osmo_stream_srv *conn = osmo_iofd_get_data(iofd);
+	if (conn->segmentation_cb2)
+		return conn->segmentation_cb2(conn, msg);
+	if (conn->segmentation_cb)
+		return conn->segmentation_cb(msg);
+	OSMO_ASSERT(0);
+	return 0;
+}
+
+/* Configure server side segmentation for the iofd */
+static void configure_srv_segmentation_cb(struct osmo_stream_srv *conn)
+{
+	/* Copy default settings */
+	struct osmo_io_ops client_ops;
+	osmo_iofd_get_ioops(conn->iofd, &client_ops);
+	/* Set segmentation cb for this client */
+	if (conn->segmentation_cb || conn->segmentation_cb2)
+		client_ops.segmentation_cb2 = stream_srv_iofd_segmentation_cb2;
+	else
+		client_ops.segmentation_cb2 = NULL;
+	osmo_iofd_set_ioops(conn->iofd, &client_ops);
+}
+
 /*! Set the segmentation callback for target osmo_stream_srv structure.
+ *  \param[in,out] conn Target Stream Server to modify
+ *  \param[in] segmentation_cb Segmentation callback to be set
+ *
+ *  DEPRECATED: Use osmo_stream_srv_set_segmentation_cb2() instead.
  *
  *  A segmentation call-back can optionally be used when a packet based protocol (like TCP) is used within a
  *  STREAM style socket that does not preserve message boundaries within the stream.  If a segmentation
@@ -981,22 +1080,41 @@ osmo_stream_srv_set_data(struct osmo_stream_srv *conn,
  *  complete single messages, and not arbitrary segments of the stream.
  *
  *  This function only works with osmo_stream_srv in osmo_io mode, created by osmo_stream_srv_create2()!
-  * The connection has to have been established prior to calling this function.
-  *
- *  \param[in,out] conn Target Stream Server to modify
- *  \param[in] segmentation_cb Segmentation callback to be set */
+ *  The connection has to have been established prior to calling this function.
+ *
+ */
 void osmo_stream_srv_set_segmentation_cb(struct osmo_stream_srv *conn,
 					 osmo_stream_srv_segmentation_cb_t segmentation_cb)
 {
 	/* Note that the following implies that iofd != NULL, since
 	 * osmo_stream_srv_create2() creates the iofd member, too */
 	OSMO_ASSERT(conn->mode == OSMO_STREAM_MODE_OSMO_IO);
-	/* Copy default settings */
-	struct osmo_io_ops conn_ops;
-	osmo_iofd_get_ioops(conn->iofd, &conn_ops);
-	/* Set segmentation cb for this connection */
-	conn_ops.segmentation_cb = segmentation_cb;
-	osmo_iofd_set_ioops(conn->iofd, &conn_ops);
+	conn->segmentation_cb = segmentation_cb;
+	conn->segmentation_cb2 = NULL;
+	configure_srv_segmentation_cb(conn);
+}
+
+/*! Set the segmentation callback for target osmo_stream_srv structure.
+ *  \param[in,out] conn Target Stream Server to modify
+ *  \param[in] segmentation_cb2 Segmentation callback to be set
+ *
+ *  A segmentation call-back can optionally be used when a packet based protocol (like TCP) is used within a
+ *  STREAM style socket that does not preserve message boundaries within the stream.  If a segmentation
+ *  call-back is given, the osmo_stream_srv library code will makes sure that the read_cb called only for
+ *  complete single messages, and not arbitrary segments of the stream.
+ *
+ *  This function only works with osmo_stream_srv in osmo_io mode, created by osmo_stream_srv_create2()!
+ *  The connection has to have been established prior to calling this function.
+ */
+void osmo_stream_srv_set_segmentation_cb2(struct osmo_stream_srv *conn,
+					 osmo_stream_srv_segmentation_cb2_t segmentation_cb2)
+{
+	/* Note that the following implies that iofd != NULL, since
+	 * osmo_stream_srv_create2() creates the iofd member, too */
+	OSMO_ASSERT(conn->mode == OSMO_STREAM_MODE_OSMO_IO);
+	conn->segmentation_cb = NULL;
+	conn->segmentation_cb2 = segmentation_cb2;
+	configure_srv_segmentation_cb(conn);
 }
 
 /*! Retrieve application private data of the stream server
@@ -1075,11 +1193,15 @@ struct osmo_stream_srv_link *osmo_stream_srv_get_master(struct osmo_stream_srv *
  *  \param[in] conn Stream Server to be destroyed */
 void osmo_stream_srv_destroy(struct osmo_stream_srv *conn)
 {
+	if (!conn)
+		return;
+
 	switch (conn->mode) {
 	case OSMO_STREAM_MODE_OSMO_FD:
 		osmo_fd_unregister(&conn->ofd);
 		close(conn->ofd.fd);
 		msgb_queue_free(&conn->tx_queue);
+		conn->tx_queue_count = 0;
 		conn->ofd.fd = -1;
 		break;
 	case OSMO_STREAM_MODE_OSMO_IO:
@@ -1111,7 +1233,12 @@ void osmo_stream_srv_send(struct osmo_stream_srv *conn, struct msgb *msg)
 
 	switch (conn->mode) {
 	case OSMO_STREAM_MODE_OSMO_FD:
-		msgb_enqueue(&conn->tx_queue, msg);
+		if (conn->tx_queue_count >= conn->srv->tx_queue_max_length) {
+			LOGSSRV(conn, LOGL_ERROR, "send: tx queue full, dropping msg!\n");
+			msgb_free(msg);
+			return;
+		}
+		msgb_enqueue_count(&conn->tx_queue, msg, &conn->tx_queue_count);
 		osmo_fd_write_enable(&conn->ofd);
 		break;
 	case OSMO_STREAM_MODE_OSMO_IO:
@@ -1181,8 +1308,12 @@ int osmo_stream_srv_recv(struct osmo_stream_srv *conn, struct msgb *msg)
 	}
 
 	if (ret < 0) {
+		if (ret == -EAGAIN) /* Received MSG_NOTIFICATION from stream_sctp_recvmsg_wrapper() */
+			return ret;
 		if (errno == EPIPE || errno == ECONNRESET)
-			LOGSSRV(conn, LOGL_ERROR, "lost connection with client\n");
+			LOGSSRV(conn, LOGL_ERROR, "lost connection with client (%d)\n", errno);
+		else
+			LOGSSRV(conn, LOGL_ERROR, "recv failed (%d)\n", errno);
 		return ret;
 	} else if (ret == 0) {
 		LOGSSRV(conn, LOGL_ERROR, "connection closed with client\n");
@@ -1198,6 +1329,7 @@ void osmo_stream_srv_clear_tx_queue(struct osmo_stream_srv *conn)
 	switch (conn->mode) {
 	case OSMO_STREAM_MODE_OSMO_FD:
 		msgb_queue_free(&conn->tx_queue);
+		conn->tx_queue_count = 0;
 		osmo_fd_write_disable(&conn->ofd);
 		break;
 	case OSMO_STREAM_MODE_OSMO_IO:
